from pathlib import Path
from typing import TYPE_CHECKING
import re

from loguru import logger

from shipaw.models.ship_types import ShipDirection

if TYPE_CHECKING:
    from shipaw.models.shipment import Shipment


def make_filename_safe(name: str) -> str:
    return re.sub(r'[ <>:"/\\|?*\x00-\x1F]', '_', name)


def get_label_stem(shipment: 'Shipment'):
    label_name = (
        f'{"Dropoff" if shipment.direction == ShipDirection.DROPOFF else "Shipping"} Label '
        f'{f"FROM {shipment.sender.address.business_name} " if shipment.sender else ""}'
        f'TO {shipment.recipient.address.business_name} '
        f'ON {shipment.shipping_date}'
    )
    return make_filename_safe(label_name)


def unused_path(filepath: Path):
    def numbered_filepath(number: int):
        return filepath if not number else filepath.with_stem(f'{filepath.stem}_{number}')

    incremented = 0
    lpath = numbered_filepath(incremented)
    while lpath.exists():
        incremented += 1
        logger.warning(f'FilePath {lpath} already exists')
        lpath = numbered_filepath(incremented)
    logger.debug(f'Using FilePath={lpath}')
    return lpath


# def get_shipment_label_path(shipment: 'Shipment') -> Path:
#     folder = get_label_folder(shipment.direction)
#     label_stem = get_label_stem(shipment)
#     label_filepath = (folder / label_stem).with_suffix('.pdf')
#     label_filepath = unused_path(label_filepath)
#     return label_filepath



from __future__ import annotations

import functools
import sys
from pathlib import Path
from typing import Literal

import loguru
from loguru import logger

"""
functions for configuring loguru
"""
CAT_COLOR_DICT = {
    'episode': 'cyan',
    'reddit': 'green',
    'backup': 'magenta',
}


def get_loguru(
    level: str = 'INFO',
    log_file: Path | None = None,
    profile: Literal['local', 'remote', 'default'] = 'local',
    color_dict: dict | None = None,
) -> logger:
    """
    Configure loguru logger

    :param level:
    :param log_file: path to log file
    :param profile: log profile to use
    :param color_dict: dictionary of log-category to colour mappings
    :return: logger
    """
    if color_dict:
        global CAT_COLOR_DICT
        CAT_COLOR_DICT = color_dict

    if profile == 'local':
        logger.info('Using local log profile')
        terminal_format = log_fmt_local_terminal
    elif profile == 'remote':
        logger.info('Using remote log profile')
        terminal_format = log_fmt_server_terminal
    else:
        raise ValueError(f'Invalid profile: {profile}')

    logger.remove()

    lvl = level.upper()
    if log_file:
        logger.add(log_file, rotation='1 day', delay=True, encoding='utf8', level=lvl)
    logger.add(sys.stderr, level=lvl, format=terminal_format)

    return logger


def log_fmt_local_terminal(record: loguru.Record) -> str:
    file_txt = f'{record["file"].path}:{record["line"]}'

    category = record['extra'].get('category', 'General')
    category_txt = f'{category.title():<9}'

    color = CAT_COLOR_DICT.get(category.lower(), 'white')
    category_txt = f'| {coloured(category_txt, color)}' if category_txt != 'General' else ''
    lvltext = f'<lvl>{record["level"]: <7}</lvl>'
    msg_txt = f'<lvl>{record["message"]}</lvl>'
    msg_txt = msg_txt.replace('{', '{{').replace('}', '}}')
    # msg_txt = f'{record['message']}'
    return f'{file_txt} | {lvltext} | {msg_txt}  \n'


def coloured(msg: str, colour: str) -> str:
    """
    Colour a message

    :param msg: message to colour
    :param colour: colour to use
    :return: coloured message
    """
    return f'<{colour}>{msg}</{colour}>'


def log_fmt_server_terminal(record) -> str:
    """
    Format for server-side logging

    :param record: log record
    :return: formatted log record
    """
    category = record['extra'].get('category', 'General')
    category = f'{category:<9}'
    colour = CAT_COLOR_DICT.get(category, 'white')

    file_line = f'{record["file"]}:{record["line"]}- {record["function"]}()'
    bot_says = f'<bold>{coloured(category, colour):<9} </bold> | {coloured(record["message"], colour)}'

    return f'<lvl>{record["level"]: <7} </lvl>| {bot_says} | {file_line}\n'


def logger_wraps(*, entries=True, exits=True, level='DEBUG') -> callable:
    """
    Decorator to log function entry and exit

    :param entries: log entry
    :param exits: log exit
    :param level: log level
    :return: decorator
    """

    def wrapper(func):
        name = func.__name__

        @functools.wraps(func)
        def wrapped(*args, **kwargs):
            logger_ = logger.opt(depth=1)
            if entries:
                logger_.log(level, f"Entering '{name}' (args={args}, kwargs={kwargs})")
            result = func(*args, **kwargs)
            if exits:
                logger_.log(level, "Exiting '{}' (result={})", name, result)
            return result

        return wrapped

    return wrapper



from __future__ import annotations

from base64 import b64encode
from pathlib import Path

from pydantic import ConfigDict, Field, model_validator

from shipaw.fapi.alerts import Alerts
from shipaw.models.base import ShipawBaseModel
from shipaw.models.label_file import get_label_stem, unused_path
from shipaw.models.shipment import Shipment
from shipaw.config import SHIPAW_SETTINGS


class ShipawTemplate(ShipawBaseModel):
    template_path: str
    context: dict = Field(default_factory=dict)

    def render_template(self, request):
        if not self.template_path:
            raise ValueError('No template_path set')
        return request.app.shipaw_settings.templates.TemplateResponse(
            request=request, name=self.template_path, context=self.context
        )


class BaseResponse(ShipawBaseModel):
    alerts: Alerts = Alerts.empty()
    data: dict | None = None
    success: bool | None = None
    status: str | None = None
    template: ShipawTemplate | None = None


class ShipmentResponse(BaseResponse):
    shipment: Shipment
    shipment_num: str | None = None
    tracking_link: str | None = None
    label_data: bytes | None = None
    label_path: Path | None = None
    tracking_links: list[str] = Field(default_factory=list)
    shipment_numbers: list[str] = Field(default_factory=list)

    @model_validator(mode='after')
    def track_links_from_tracking_link(self):
        if self.tracking_link and not self.tracking_links:
            self.tracking_links = [self.tracking_link]
        if self.shipment_num and not self.shipment_numbers:
            self.shipment_numbers = [self.shipment_num]
        return self

    model_config = ConfigDict(json_encoders={bytes: lambda v: b64encode(v).decode('utf-8') if v else None})

    @model_validator(mode='after')
    def get_label_path(self):
        if self.label_path is None:
            folder = SHIPAW_SETTINGS.label_dir / self.shipment.direction
            label_stem = get_label_stem(self.shipment)
            label_filepath = (folder / label_stem).with_suffix('.pdf')
            self.label_path = unused_path(label_filepath)
        return self

    # async def write_label_file(self):
    #     await array_write_label_content(self.label_data, self.label_path)


class ShipawTemplateResponse(BaseResponse):
    template: ShipawTemplate



from pathlib import Path

from fastapi import APIRouter, Body, Depends
from loguru import logger
from starlette.requests import Request
from starlette.responses import JSONResponse

from shipaw.fapi.ui_funcs import make_nice_str
from shipaw.fapi.alerts import Alert, AlertType, Alerts
from shipaw.fapi.backend import maybe_alert_apc, try_book_shipment, try_get_write_label
from shipaw.fapi.form_data import provider_from_form, shipment_request_form, shipment_request_form_json
from shipaw.fapi.requests import ShipmentRequest
from shipaw.fapi.responses import ShipawTemplate, ShipawTemplateResponse
from shipaw.logging import log_obj
from shipaw.models.shipment import Shipment
from shipaw.providers.registry import PROVIDER_REGISTER

router = APIRouter()


# router.mount('/static', StaticFiles(directory=str(SHIPAW_SETTINGS.static_dir)), name='static')


def get_version():
    from importlib.metadata import version, PackageNotFoundError

    try:
        return version('shipaw')
    except PackageNotFoundError:
        return 'unknown'


async def notify_version(request):
    alerts = Alerts.empty()
    sandbox = request.app.shipaw_settings.shipper_live == False
    if sandbox:
        live_msg = 'Test Mode - No Shipments will be booked'
        notification_type = AlertType.NOTIFICATION
    else:
        live_msg = 'Live Mode - Real Shipments will be booked'
        notification_type = AlertType.WARNING

    msg = f'Shipaw Version {get_version()} is in {live_msg}'
    logger.warning(msg)
    alerts += Alert(message=msg, type=notification_type)
    return alerts


def notify_dev() -> Alerts:
    alerts = Alerts.empty()
    if any(['prdev' in str(_).lower() for _ in Path(__file__).parents]):
        msg = '"prdev" in cwd tree - BETA MODE - This is a development version'
        logger.warning(msg)
        alerts += Alert(message=msg, type=AlertType.WARNING)
    return alerts


@router.post('/shipping_form', response_model=ShipawTemplateResponse)
async def shipping_form_api(request: Request, shipment: Shipment = Body(...)) -> ShipawTemplateResponse:
    log_obj(shipment, 'Shipment received at /ship_form:')

    alerts: Alerts = request.app.alerts
    alerts += notify_dev()
    alerts += await notify_version(request)

    tmplt = ShipawTemplate(template_path='shipping_form_container.html', context={'shipment': shipment})
    return ShipawTemplateResponse(template=tmplt, alerts=alerts)


@router.post('/order_summary', response_model=ShipawTemplateResponse)
async def order_summary_api(
    request: Request,
    shipment_request: ShipmentRequest = Depends(shipment_request_form),
) -> ShipawTemplateResponse:
    log_obj(shipment_request, 'ShipmentRequest received at shipaw/order_summary:')
    context = {'shipment_request': shipment_request}

    # check phone number
    # alerts = await maybe_alert_phone_number(shipment_request.shipment.remote_full_contact.contact.mobile_phone)
    # check not apc + dropoff
    alerts = await maybe_alert_apc(shipment_request)
    return ShipawTemplateResponse(
        template=ShipawTemplate(template_path='/order_summary.html', context=context),
        alerts=alerts,
    )


@router.post('/order_results', response_model=ShipawTemplateResponse)
async def order_results_api(
    request: Request,
    shipment_request: ShipmentRequest = Depends(shipment_request_form_json),
) -> ShipawTemplateResponse:
    shipment_response = await try_book_shipment(shipment_request)
    # await try_get_write_label(shipment_request, shipment_response)

    if shipment_response.alerts.errors:
        return await errored_shipment(shipment_response)

    log_obj(shipment_response, 'Shipment Booked')
    await try_get_write_label(shipment_request, shipment_response)

    if hasattr(request.app, 'callback'):
        await request.app.callback(shipment_request, shipment_response)

    shipment_response.template = ShipawTemplate(
        template_path='/order_results.html',
        context={'shipment_request': shipment_request, 'response': shipment_response},
    )
    return ShipawTemplateResponse.model_validate(shipment_response, from_attributes=True)


async def errored_shipment(shipment_response):
    log_obj(shipment_response.alerts, 'Errors booking shipment:')
    alerts = shipment_response.alerts
    shipment_response.template = ShipawTemplate(
        template_path='/alerts.html',
        context={'alerts': alerts},
    )
    return ShipawTemplateResponse.model_validate(shipment_response, from_attributes=True)


@router.get('/providers', response_class=JSONResponse)
async def providers():
    provider_response = {_.title(): _ for _ in PROVIDER_REGISTER.keys()}
    return JSONResponse(provider_response)


@router.get('/provider_directions/{provider_name}', response_class=JSONResponse)
async def provider_directions(provider_name: str):
    provider = await provider_from_form(provider_name)
    directions = provider.valid_directions
    dir_dict = {_.value: _.value for _ in directions}
    return JSONResponse(dir_dict)


@router.get('/provider_direction_services/{provider_name}/{direction}', response_class=JSONResponse)
async def provider_direction_services(provider_name: str, direction: str):
    provider = await provider_from_form(provider_name)
    list_of_str_enums = provider.valid_direction_services.get(direction, list())
    res_dir = {make_nice_str(_.name): _.value for _ in list_of_str_enums}
    return JSONResponse(res_dir)



from pydantic import AliasGenerator, BaseModel, ConfigDict
from pydantic.alias_generators import to_pascal


class ShipawBaseModel(BaseModel):
    model_config = ConfigDict(
        alias_generator=AliasGenerator(
            alias=to_pascal,
        ),
        populate_by_name=True,
        use_enum_values=True,
    )



from __future__ import annotations

from dataclasses import dataclass
from pathlib import Path

import pythoncom
from loguru import logger
from win32com.client import Dispatch

from shipaw.config import SHIPAW_SETTINGS
from shipaw.fapi.requests import ShipmentRequest


@dataclass
class Email:
    to_address: str
    subject: str
    body: str
    attachment_paths: list[Path] | None = None

    def __post_init__(self):
        if self.attachment_paths is None:
            self.attachment_paths = []

    def send(self, sender: OutlookHandler) -> None:
        sender.create_open_email(self)


class OutlookHandler:
    """
    Email handler for Outlook (ripped from pawsupport where it has a superclass and siblings for Gmail etc)
    """

    @staticmethod
    def create_open_email(email: Email, html: bool = False):
        """
        Send email via Outlook

        :param email: Email object
        :param html: format email from html input
        :return: None
        """
        try:
            pythoncom.CoInitialize()

            outlook = Dispatch('outlook.application')
            mail = outlook.CreateItem(0)
            mail.To = email.to_address
            mail.Subject = email.subject
            if html:
                mail.HtmlBody = email.body
            else:
                mail.Body = email.body

            for att_path in email.attachment_paths:
                mail.Attachments.Add(str(att_path))
                print('Added attachment')
            mail.Display()
        except Exception as e:
            logger.exception(f'Failed to send email with error: {e}')
            raise ValueError(f'{e.args[0]}')
        finally:
            pythoncom.CoUninitialize()


async def subject(*, invoice_num: str | None = None, missing: bool = False, label: bool = False):
    return (
        f'Amherst Radios'
        f'{f"- Invoice {invoice_num} Attached" if invoice_num else ""} '
        f'{"- We Are Missing Kit" if missing else ""} '
        f'{"- Shipping Label Attached" if label else ""}'
    )


async def send_label_email(shipment_request: ShipmentRequest, label_path: Path):
    body = SHIPAW_SETTINGS.templates.get_template('email_snips/label_email.html').render(
        label=label_path,
        shipment_request=shipment_request,
        home_business_name=SHIPAW_SETTINGS.business_name,
    )
    email = Email(
        to_address=shipment_request.shipment.remote_full_contact.contact.email_address,
        subject='Amherst Radios Shipping - Shipping Label Attached',
        body=body,
        attachment_paths=[label_path],
    )

    OutlookHandler.create_open_email(email, html=True)



from __future__ import annotations

from pathlib import Path

from apc_hypaship.error import apc_http_status_alerts
from httpx import HTTPStatusError
from loguru import logger
from pawdf.array_pdf.array_p import on_a4

from shipaw.fapi.alerts import Alert, AlertType, Alerts
from shipaw.fapi.requests import ShipmentRequest
from shipaw.fapi.responses import ShipmentResponse
from shipaw.models.ship_types import ShipDirection

from shipaw.providers.provider_abc import ProviderName


async def try_book_shipment(shipment_request: ShipmentRequest) -> ShipmentResponse:
    shipment_response = ShipmentResponse(shipment=shipment_request.shipment)
    try:
        shipment_response = shipment_request.provider.book_shipment_agnostic(shipment_request)

    except HTTPStatusError as e:
        await maybe_apc_response_error(e, shipment_request, shipment_response)

    except Exception as e:
        logger.exception(f'Error booking shipment: {e}')
        shipment_response.alerts += Alert.from_exception(e)

    return shipment_response


async def try_get_write_label(request: ShipmentRequest, response: ShipmentResponse):
    if not response.label_data:
        await try_get_label_data(request, response)
    # await try_write_label(response)
    await try_write_label2(response)


async def try_get_label_data(request: ShipmentRequest, response: ShipmentResponse) -> None:
    try:
        if response.label_data is not None:
            logger.info('Label data already present, not fetching')
        else:
            if response.shipment_num:
                response.label_data = await request.provider.wait_fetch_label_async(response.shipment_num)
            else:
                logger.warning('No shipment number to fetch label data')

    except HTTPStatusError as e:
        await maybe_apc_response_error(e, request, response)

    except Exception as e:
        logger.exception('Error getting label data')
        response.alerts += Alert.from_exception(e)


# async def try_write_label(response: ShipmentResponse):
#     try:
#         await response.write_label_file()
#     except Exception as e:
#         logger.exception(f'Error writing label file: {e}')
#         response.alerts += Alert.from_exception(e)


async def try_write_label2(response: ShipmentResponse):
    try:
        label_content = response.label_data
        label_path = response.label_path
        await array_write_label_content(label_content, label_path)

    except Exception as e:
        logger.exception(f'Error writing label file: {e}')
        response.alerts += Alert.from_exception(e)


async def maybe_apc_response_error(e: HTTPStatusError, shipment_request, shipment_response):
    if shipment_request.provider_name == ProviderName.APC:
        for alert in await apc_http_status_alerts(e):
            shipment_response.alerts += alert
    else:
        logger.exception(e)
        shipment_response.alerts += Alert.from_exception(e)


async def maybe_alert_apc(shipment_request):
    alerts = Alerts.empty()
    if (
        shipment_request.provider_name == ProviderName.APC
        and shipment_request.shipment.direction == ShipDirection.DROPOFF
    ):
        alerts += Alert(
            message='APC does not support drop-off shipments - please select Outbound or Inbound Collection',
            type=AlertType.ERROR,
        )
    return alerts


async def array_write_label_content(label_content: bytes, label_path: Path):
    og_size_path = label_path.parent / 'original_size' / label_path.name
    og_size_path.parent.mkdir(parents=True, exist_ok=True)
    og_size_path.write_bytes(label_content)
    logger.info(f'Resizing {og_size_path} to A4 at {label_path}')
    on_a4(input_file=og_size_path, output_file=label_path)
    logger.info(f'Wrote label to {label_path}')
